;;; g-completion.el --- function and variable for Emacs
;;
;; Copyright (c) ?????
;;
;; This file is for C and C++ code navigation.
;;
;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;; 
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;; 
;; You should have received a copy of the GNU General Public License
;; from <http://www.gnu.org/licenses/>.
;;
;; version: 0.1

(define-derived-mode g-si-mode
  c-mode "g-si")

(defvar g-si-flag nil)
(defvar g-si-temp nil)
(defvar g-si-point nil)
(defvar g-si-file nil)
(defvar g-si-line nil)
(defvar g-si-char nil)
(defvar g-si-semantic nil)
(defvar g-si-buffer nil)
(defvar g-si-sem-flag nil)  ;;1 means find from buffer, 2 means find from file
(defvar g-si-back-buffer nil)
(defvar g-si-back-point nil)
(defvar g-si-have-find nil)
(defvar g-si-file-exist-stack nil)
(defvar g-si-file-stack nil)
(defvar g-si-point-stack nil)
(defvar g-si-open-stack nil)

(define-key g-si-mode-map [mouse-1] 'g-si-mouse-1)

(defun g-si-mouse-1()
(interactive)
 (cond
  ;; 1 means find one define from semantic
  ((= 1 g-si-flag)
   (message "from-semantic")
   (unless (equal g-si-have-find nil)
     (g-si-open-file-1)
     (g-si-push-mark)
   ))
  ;; 2 means find one define from gtags
  ((= 2 g-si-flag)
   (message "from-gtags-one")
   (g-si-open-file-2)
   (g-si-push-mark)
   )
  ;; 3 means find lots of result from gtags
  ((= 3 g-si-flag)
   (g-si-open-file-3)
   (g-si-push-mark)
   (message "from-gtags-many")
   )
  )
)

(defun g-si-open-file-1()
  (cond ((= g-si-sem-flag 1)
         (setq g-si-file-exist-stack (cons t g-si-file-exist-stack))
         (switch-to-buffer-other-window g-si-buffer)
         (goto-char g-si-char))
        ((= g-si-sem-flag 2)
         (if (get-file-buffer g-si-file)
             (setq g-si-file-exist-stack (cons t g-si-file-exist-stack))
           (setq g-si-file-exist-stack (cons nil g-si-file-exist-stack)))
         (switch-to-buffer-other-window (find-file-noselect g-si-file))
         (goto-char g-si-char))))

(defun g-si-open-file-2()
  (if (get-file-buffer g-si-file)
      (setq g-si-file-exist-stack (cons t g-si-file-exist-stack))
    (setq g-si-file-exist-stack (cons nil g-si-file-exist-stack)))
  (switch-to-buffer-other-window (find-file-noselect g-si-file))
  (goto-line g-si-line))

(defun g-si-open-file-3()
  (gtags-select-it1)
  (if (get-file-buffer g-si-file)
      (setq g-si-file-exist-stack (cons t g-si-file-exist-stack))
    (setq g-si-file-exist-stack (cons nil g-si-file-exist-stack)))
  (switch-to-buffer-other-window (find-file-noselect g-si-file))
  (goto-line g-si-line))

(defun g-si-enable()
(interactive)
(get-buffer-create "aaa")
(switch-to-buffer-other-window "aaa")
(save-current-buffer
  (set-buffer "aaa")
  (g-si-mode)))

(defun g-si-show-define(point)
(interactive "d")
(setq g-si-point point)
(setq g-si-back-buffer (current-buffer))
(setq g-si-back-point point)
(run-with-idle-timer 0.25 nil 'g-si-find))

(defun g-si-directly-show-define()
(setq g-si-point (point))
(setq g-si-back-buffer (current-buffer))
(setq g-si-back-point point)
(g-si-find)
)

(defun g-si-find()
(let (tagname)
  (setq g-si-have-find nil)
  (setq tagname (gtags-current-token))
  (gtags-goto-tag1 tagname)
  (if (= 1 g-si-semantic)
      (progn
        (setq g-si-semantic nil)
        (g-si-semantic-1 g-si-point)))))

(defun g-si-push-mark()
;;(message "g-si-push-mark")
(setq g-si-file-stack (cons (buffer-file-name g-si-back-buffer) g-si-file-stack))
(if (equal g-si-file-stack nil)
    (message "f-stack-empty")
  (message "f-stack-push"))
(setq g-si-point-stack (cons g-si-back-point g-si-point-stack))
(setq g-si-open-stack (cons (buffer-file-name (current-buffer)) g-si-open-stack))
)

(defun g-si-pop-mark()
(interactive)
(if (not g-si-file-stack) nil
  (let (point file exist)
    (setq exist (car g-si-file-exist-stack))
    (setq g-si-file-exist-stack (cdr g-si-file-exist-stack))
    (if (not exist)
        (progn
          (setq file (car g-si-open-stack))
          (setq g-si-open-stack (cdr g-si-open-stack))
          (kill-buffer (get-file-buffer file))))
    (setq file (car g-si-file-stack))
    (setq g-si-file-stack (cdr g-si-file-stack))
    (setq point (car g-si-point-stack))
    (setq g-si-point-stack (cdr g-si-point-stack))
    (switch-to-buffer (get-file-buffer file))
    (goto-char point)
    (g-si-directly-show-define)))
)

(defun gtags-goto-tag1 (tagname)
   (let (lines line file buffer window)
     (setq save (current-buffer))
     (save-current-buffer
       (set-buffer "aaa")
       (erase-buffer))
     (call-process "global" nil "aaa" nil "-x" tagname)
     (save-current-buffer
       (set-buffer "aaa")
     (goto-char (point-min))
     (setq lines (count-lines (point-min) (point-max)))
     (cond
      ((= 0 lines)
       (setq g-si-semantic 1))
      ((= 1 lines)
       (setq g-si-flag 2)
       (gtags-select-it1))
      (t
       (setq g-si-flag 3)
       (message "gtags find many"))))
     (if (= 1 lines)
         (progn
           (save-current-buffer
             (set-buffer "aaa")
             (erase-buffer)
             (insert-file-contents g-si-file)
             (with-selected-window
                 (setq window (car (get-buffer-window-list "aaa")))
               (select-window window)
               (goto-line g-si-line)))))
     )
  )

(defun gtags-select-it1 ()
    (beginning-of-line)
    (if (not (looking-at "[^ \t]+[ \t]+\\([0-9]+\\)[ \t]\\([^ \t]+\\)[ \t]"))
        (message "gtags error")
      (setq g-si-line (string-to-number (gtags-match-string 1)))
      (save-current-buffer
        (setq g-si-file (gtags-match-string 2))
        (set-buffer g-si-back-buffer)
        (setq g-si-file (expand-file-name g-si-file)))
      (message "line:%d file:%s" g-si-line g-si-file)
      
))

(defun g-si-semantic-1 (point)
  "Jump to the tag referred to by the code at POINT.
Uses `semantic-analyze-current-context' output to identify an accurate
origin of the code at point."
  (interactive "d")
  (message "g-si-semantic-1")
  (let* ((ctxt (semantic-analyze-current-context point))
	 (pf (and ctxt (reverse (oref ctxt prefix))))
	 ;; In the analyzer context, the PREFIX is the list of items
	 ;; that makes up the code context at point.  Thus the c++ code
	 ;; this.that().theothe
	 ;; would make a list:
	 ;; ( ("this" variable ..) ("that" function ...) "theothe")
	 ;; Where the first two elements are the semantic tags of the prefix.
	 ;;
	 ;; PF is the reverse of this list.  If the first item is a string,
	 ;; then it is an incomplete symbol, thus we pick the second.
	 ;; The second cannot be a string, as that would have been an error.
	 (first (car pf))
	 (second (nth 1 pf))
	 )
    (cond
     ((semantic-tag-p first)
      ;; We have a match.  Just go there.
      (g-si-semantic-2 first))
     
;;      ((semantic-tag-p second)
;;        ;; Because FIRST failed, we should visit our second tag.
;;        ;; HOWEVER, the tag we actually want that was only an unfound
;;        ;; string may be related to some take in the datatype that belongs
;;        ;; to SECOND.  Thus, instead of visiting second directly, we
;;        ;; can offer to find the type of SECOND, and go there.
;;        (let ((secondclass (car (reverse (oref ctxt prefixtypes)))))
;;  	(cond
;;  	 ((and (semantic-tag-with-position-p secondclass)
;;  	       (y-or-n-p (format "Could not find `%s'.  Jump to %s? "
;;  				 first (semantic-tag-name secondclass))))
;;  	  (g-si-semantic-2 secondclass)
;;  	  )
;;  	 ;; If we missed out on the class of the second item, then
;;  	 ;; just visit SECOND.
;;  	 ((and (semantic-tag-p second)
;;  	       (y-or-n-p (format "Could not find `%s'.  Jump to %s? "
;;  				 first (semantic-tag-name second))))
;;  	  (g-si-semantic-2 second)
;;  	  ))))

;;      ((semantic-tag-of-class-p (semantic-current-tag) 'include)
;;       ;; Just borrow this cool fcn.
;;       (semantic-decoration-include-visit)
;;       )

     (t
      (error "Could not find suitable jump point for %s"
	     first))
     )))

(defun g-si-semantic-2 (dest)
  "Jump to DEST, a Semantic tag.
This helper manages the mark, buffer switching, and pulsing."
  ;; We have a tag, but in C++, we usually get a prototype instead
  ;; because of header files.  Lets try to find the actual
  ;; implementaion instead.
  (when (semantic-tag-prototype-p dest)
    (let* ((refs (semantic-analyze-tag-references dest))
	   (impl (semantic-analyze-refs-impl refs t))
	   )
      (when impl (setq dest (car impl)))))

  ;; Make sure we have a place to go...
  (if (not (and (or (semantic-tag-with-position-p dest)
		    (semantic-tag-get-attribute dest :line))
		(semantic-tag-file-name dest)))
      (error "Tag %s has no buffer information"
	     (semantic-format-tag-name dest)))

  ;; Once we have the tag, we can jump to it.  Here
  ;; are the key bits to the jump:

  ;; 1) Push the mark, so you can pop global mark back, or
  ;;    use semantic-mru-bookmark mode to do so.
;;   (push-mark)
;;   (when (fboundp 'push-tag-mark)
;;     (push-tag-mark))
  ;; 2) Visits the tag.
  (g-si-semantic-3 dest)
  ;; 3) go-to-tag doesn't switch the buffer in the current window,
  ;;    so it is like find-file-noselect.  Bring it forward.
;;  (switch-to-buffer (current-buffer))
  ;; 4) Fancy pulsing.
;;  (pulse-momentary-highlight-one-line (point))
  )

(defun g-si-semantic-3(tag)
   (cond ((semantic-tag-in-buffer-p tag)
	  ;; We have a linked tag, go to that buffer.
          (setq g-si-buffer (semantic-tag-buffer tag))
	  (set-buffer g-si-buffer)
          (setq g-si-temp (buffer-substring-no-properties (point-min) (point-max)))
          (set-buffer "aaa")
          (erase-buffer)
          (setq g-si-sem-flag 1)
          (insert g-si-temp))
	 ((semantic-tag-file-name tag)
	  ;; If it didn't have a buffer, but does have a file
	  ;; name, then we need to get to that file so the tag
	  ;; location is made accurate.
          (setq g-si-file (semantic-tag-file-name tag))
          (setq g-si-file (expand-file-name g-si-file))
          (set-buffer "aaa")
          (erase-buffer)
          (setq g-si-sem-flag 2)
          ;; (message "sem-find-file")
          (insert-file-contents g-si-file))
	 (t
	  ;; Well, just assume things are in the current buffer.
	  nil
	  ))
   ;; We should be in the correct buffer now, try and figure out
   ;; where the tag is.
   (let (window)
     (with-selected-window
         (setq window (car (get-buffer-window-list "aaa")))
       (select-window window)
       (cond ((semantic-tag-with-position-p tag)
              ;; If it's a number, go there
              (message "sem-num")
              (setq g-si-flag 1)
              (setq g-si-have-find t)
              (setq g-si-char (semantic-tag-start tag))
              (goto-char g-si-char))
             ((semantic-tag-get-attribute tag :line)
              ;; The tag has a line number in it.  Go there.
              (setq g-si-flag 1)
              (message "sem-line")
              (setq g-si-have-find t)
              (goto-line (semantic-tag-get-attribute tag :line)))
             (t
              ;; Take a guess that the tag has a unique name, and just
              ;; search for it from the beginning of the buffer.
              (message "sem-char")
              (goto-char (point-min))
              (re-search-forward (semantic-tag-name tag) nil t)))))
  )

(defun g-si-show-ref(point)
(interactive "d")
(setq g-si-point point)
(setq g-si-back-buffer (current-buffer))
(setq g-si-back-point point)
(g-si-find-ref))

(defun g-si-find-ref()
(let (tagname)
  (setq g-si-have-find nil)
  (setq tagname (gtags-current-token))
  (if (equal tagname nil)
      (yank)
    (gtags-find-all tagname))))


(defun gtags-find-all (tagname)
   (let (lines line file buffer window)
     (setq save (current-buffer))
     (save-current-buffer
       (set-buffer "aaa")
       (erase-buffer))
     (call-process "global" nil "aaa" nil "-xr" tagname)
     (call-process "global" nil "aaa" nil "-xs" tagname)
     (save-current-buffer
       (set-buffer "aaa")
     (goto-char (point-min))
     (setq lines (count-lines (point-min) (point-max)))
     (cond
      ((= 0 lines)
       (message "global find nothing"))
      ((= 1 lines)
       (setq g-si-flag 2)
       (gtags-select-it1))
      (t
       (setq g-si-flag 3)
       (message "gtags find many"))))
     (if (= 1 lines)
         (progn
           (save-current-buffer
             (set-buffer "aaa")
             (erase-buffer)
             (insert-file-contents g-si-file)
             (with-selected-window
                 (setq window (car (get-buffer-window-list "aaa")))
               (select-window window)
               (goto-line g-si-line)))))
     )
  )



(provide 'g-si)

;;; g-si.el ends here
